@@ -2,7 +2,7 @@ use strict;
 use warnings;
 package Plack::Middleware::Auth::Form;
 BEGIN {
-  $Plack::Middleware::Auth::Form::VERSION = '0.010';
+  $Plack::Middleware::Auth::Form::VERSION = '0.011';
 }
 
 use parent qw/Plack::Middleware/;
@@ -26,12 +26,11 @@ sub prepare_app {
 sub call {
     my($self, $env) = @_;
     my $path = $env->{PATH_INFO};
-
+    
     if( $env->{'psgix.session'}{remember} ){
         if( $path ne '/logout' ){
             $env->{'psgix.session.options'}{expires} = time + 60 * 60 * 24 * 30;
         }
-        delete $env->{'psgix.session'}{remember};
     }
 
     if( $path eq '/login' ){
@@ -59,26 +58,31 @@ sub _login {
         ];
     }
     my $params = Plack::Request->new( $env )->parameters;
-    if( defined $env->{user} ){
-        return 'Already logged in';
-    }
-    elsif( $env->{REQUEST_METHOD} eq 'POST' ){
+    if( $env->{REQUEST_METHOD} eq 'POST' ){
         my $user_id;
         my $auth_result = $self->authenticator->( $params->get( 'username' ), $params->get( 'password' ), $env );
+        my $redir_to;
         if( ref $auth_result ){
             $login_error = $auth_result->{error};
             $user_id = $auth_result->{user_id};
+            $redir_to = $auth_result->{redir_to};
         }
         else{
             $login_error = 'Wrong username or password' if !$auth_result;
             $user_id = $params->get( 'username' );
+            delete $env->{'psgix.session'}{user_id};
+            delete $env->{'psgix.session'}{remember};
         }
         if( !$login_error ){
+            $env->{'psgix.session.options'}->{change_id}++;
             $env->{'psgix.session'}{user_id} = $user_id;
-            $env->{'psgix.session'}{remember} = 1 if $params->get( 'remember' );
-            my $redir_to = delete $env->{'psgix.session'}{redir_to};
-            $redir_to = '/' if 
-                URI->new( $redir_to )->path eq $env->{PATH_INFO};
+            $env->{'psgix.session'}{remember} = ($params->get( 'remember' ) ? 1 : 0);
+            my $tmp_redir = delete $env->{'psgix.session'}{redir_to};
+            if( !defined($redir_to) ){
+                $redir_to = $tmp_redir;
+                $redir_to = '/' if 
+                    URI->new( $redir_to )->path eq $env->{PATH_INFO};
+            }
             return [ 
                 302, 
                 [ Location => $redir_to ],
@@ -86,6 +90,13 @@ sub _login {
             ];
         }
     }
+    elsif( defined $env->{'psgix.session'}{user_id} ){
+        return [ 
+            200, 
+            [ 'Content-Type' => 'text/html', ],
+            [ $self->_wrap_body( 'Already logged in' ) ]
+        ];
+    }
     $env->{'psgix.session'}{redir_to} ||= $env->{HTTP_REFERER} || '/';
     my $form = $self->_render_form( 
         username => $params->get( 'username' ), 
@@ -130,6 +141,7 @@ sub _logout {
     my($self, $env) = @_;
     if( $env->{REQUEST_METHOD} eq 'POST' ){
         delete $env->{'psgix.session'}{user_id};
+        delete $env->{'psgix.session'}{remember};
     }
     return [ 
         303, 
@@ -157,7 +169,7 @@ Plack::Middleware::Auth::Form - Form Based Authentication for Plack (think L<Cat
 
 =head1 VERSION
 
-version 0.010
+version 0.011
 
 =head1 SYNOPSIS
 
@@ -182,7 +194,14 @@ logouts the user (only on a POST) and redirects him to C<after_logout> or C</>.
 =back
 
 After a succesful login the user is redirected back to url identified by 
-the C<redir_to> session parameter.  It also sets that session parameter from
+the C<redir_to> session parameter.  You can set it using the psgix session
+hash:
+
+    $env->{'psgix.session'}{redir_to} = '/some/page';
+
+see L<Plack::Middleware::Session> for more explanations.
+
+It also sets that session parameter from
 C<< $env->{HTTP_REFERER} >> if it is not set or to C</> if even that is not available.
 The username (or id) is saved to C<user_id> session parameter, if you want
 to save an id different from the username - then you need to return
@@ -209,9 +228,10 @@ backends for L<Authen::Simple> is perfect to use:
   use Authen::Simple::LDAP;
   enable "Auth::Form", authenticator => Authen::Simple::LDAP->new(...);
 
-The callback can also return a hashref with two optional fields
-C<error> - the reason for the failure and C<user_id> - the user id
-to be saved in the session instead of the username.
+The callback can also return a hashref with three optional fields
+C<error> - the reason for the failure, C<user_id> - the user id
+to be saved in the session instead of the username, and C<redir_to> - a user
+defined redirection page.
 
 =item no_login_page
 
@@ -246,6 +266,10 @@ L<Plack::Middleware::Auth::Basic>.
 =head1 CONTRIBUTORS
 
 Tokuhiro Matsuno
+chromatic
+hayajo
+Kaare Rasmussen
+Oliver Paukstadt
 
 =head1 AUTHOR
 
